
/*--------------------------------------------------------------------*/
/*--- Operations on the list of exclusions for in-code data        ---*/
/*---                                                              ---*/
/*---                                         rise_exclusionList.c ---*/
/*--------------------------------------------------------------------*/

/*
   This file is part of RISE, an instruction set randomizer for security,
   built on top of Valgrind, an extensible x86 protected-mode
   emulator for monitoring program execution on x86-Unixes.

   RISE:
   Copyright (C) 2002-2003 Gabriela Barrantes
   	gbarrant@ieee.org

   Valgrind:
   Copyright (C) 2000-2003 Julian Seward 
      jseward@acm.org

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.

   The GNU General Public License is contained in the file COPYING.
*/
#include "vg_include.h"

//Some helper functions duplicated or slightly modified from coregrind/vg_procselfmaps

Int hexdigit ( Char c )
{
	   if (c >= '0' && c <= '9') return (Int)(c - '0');
	      if (c >= 'a' && c <= 'f') return 10 + (Int)(c - 'a');
	         if (c >= 'A' && c <= 'F') return 10 + (Int)(c - 'A');
		    return -1;
}

Int readhex ( Char* buf, UInt limit, UInt* val )
{
   Int hed;
   UInt offset = 0;
   *val = 0;
   
   while ( (((Addr)buf+offset)<limit) && ((hed=hexdigit(buf[offset])) >= 0)) {
       *val = (*val << 4) + hed;
        offset++;
    }
  
   if ( (((Addr)buf+offset)==limit) || ((buf[offset]!=',') && (buf[offset]!='\n')))
	   return 0;
   else
	   return offset;
	
  	 
}
//End of helper fns from procselfmaps


Int readName(Char *buf,UInt limit) {
   UInt offset = 0;
   while ( (((Addr)buf+offset)<limit) && (buf[offset]!=',') ) {
        offset++;
    }
   if (((Addr) buf+offset)==limit)
	   return 0;
   else
	   return offset;
}
	
typedef
  struct _excAdd {
	Addr eAdd;
	struct _excAdd *next;
} oneAddress;

typedef
   struct
      _libEL {
	char *libname;
	oneAddress *aList;
	struct _libEL *next;
      } libExcList;

libExcList * headEL = NULL;

void printListRead(void) {
	libExcList *trav;
	oneAddress *adrs;

	trav=headEL;
	while (trav !=NULL) {
		VG_(printf)("%s\n",trav->libname);
		adrs=trav->aList;
		while (adrs != NULL) {
			VG_(printf)("%x\n",adrs->eAdd);
			adrs = adrs->next;
		}
		trav=trav->next;
	}
}

void RISE_(load_exclusion_list)(char * filename) {

	Int fd,i,opResult;
	Addr elImage;
	struct vki_stat stat_buf;
	UInt sizeFile;
	Char *curC;
	oneAddress *newAdd;
	libExcList *newLib;
	UInt numDigits;
	UInt Number;
	

	i = VG_(stat)(filename, &stat_buf);
	if (i != 0) {
		VG_(message)(Vg_UserMsg,"Cannot determine size of exclusion list file\n");
		VG_(exit)(1);
	}
	sizeFile = stat_buf.st_size;


   	fd = VG_(open)(filename, VKI_O_RDONLY, 0);
      	if (fd == -1) {
	            VG_(message)(Vg_UserMsg,"Can't open exclusion list");
		    VG_(exit)(1);
	}

	elImage = (Addr)VG_(mmap)( NULL, sizeFile,
	                          VKI_PROT_READ, VKI_MAP_PRIVATE, fd, 0 );
        if (elImage == ((Addr)(-1))) {
		VG_(message)(Vg_UserMsg, "mmap failed on %s", filename );
	       	VG_(close)(fd);
		VG_(exit)(1);
        }
	
	VG_(close)(fd);

	//The image of the exclusion file is mapped to memory -- read it
	//Format:
	//library-name1,hex-no1,hex-no2,...,hex-non1\n
	//library-name2,hex-no1,...
	//...
	//library-namem,...,hex-nonm
	
	curC = (Char *)elImage;
	i=0;

	while (True) {

		//Create new library element
		newLib = (libExcList *)VG_(arena_malloc)(VG_AR_CORE, sizeof(libExcList));
		if (newLib ==NULL) {
                        VG_(message)(Vg_UserMsg,"Error allocating memory for library info\n");
			VG_(exit)(1);
		}
		
		newLib->next = NULL;
		newLib->aList = NULL;
		newLib->libname = NULL;

		//Scans the library name to determine the number of characters
		opResult = readName(curC,elImage+sizeFile);

		if (opResult == 0) {
	        	VG_(message)(Vg_UserMsg,"Error reading library name\n");
			VG_(exit)(1);
		}
		
		//Allocates enough memory for the name plus 0
		newLib->libname = (Char *) VG_(arena_malloc)(VG_AR_CORE,opResult+1 );
		if (newLib->libname ==NULL) {
                        VG_(message)(Vg_UserMsg,"Error allocating memory for library info\n");
			VG_(exit)(1);
		}

		//Copies the library name to the new location
		VG_(strncpy)(newLib->libname,curC,opResult);
		newLib->libname[opResult]=0;

		//Updates the traverse pointer in the map
		//At each traverse pointer update we must check if we are out 
		//of boundaries
		curC+=opResult;
		if ((Addr) curC == ((Addr)elImage + sizeFile)) {
        		VG_(message)(Vg_UserMsg,"Error in exclusion file format\n");
	         	VG_(exit)(1);
		}
		

		while (True) {

			//Checks if the separator character is there
			if (curC[0]!=','){
        			VG_(message)(Vg_UserMsg,"Error in exclusion file format\n");
	         		VG_(exit)(1);
			}

			curC++;
			if ((Addr) curC == ((Addr)elImage + sizeFile)) {
        			VG_(message)(Vg_UserMsg,"Error in exclusion file format\n");
	         		VG_(exit)(1);
			}
			
			//Obtains the hex number
			numDigits = readhex(curC,elImage+sizeFile,&Number);

			if (numDigits == 0) {
        			VG_(message)(Vg_UserMsg,"Not a number in library offsets\n");
	         		VG_(exit)(1);
			}

			//Adds the number to the list for the library
			newAdd = (oneAddress *)VG_(arena_malloc)(VG_AR_CORE, sizeof(oneAddress));
			if (newAdd == NULL) {
                        	VG_(message)(Vg_UserMsg,"Error allocating memory for library info\n");
				VG_(exit)(1);
			}
			
			newAdd->eAdd = Number;
			newAdd->next = NULL;

			if (newLib->aList == NULL)
			 	newLib->aList = newAdd;
			else {
				newAdd->next = newLib->aList;
				newLib->aList = newAdd;
			}
				
			curC+=numDigits;
			if ((Addr) curC == ((Addr) elImage + sizeFile)) {
        			VG_(message)(Vg_UserMsg,"Error in exclusion file format\n");
	         		VG_(exit)(1);
			}
		
			//If \n detected finishes
			if ((curC[0] == '\n') ||  ((Addr) curC == ((Addr) elImage + sizeFile)))
				break;
		
		}

		curC++;
	
		//Connects the new library info to the list
	
		if (headEL != NULL) 
			newLib->next = headEL;
		headEL = newLib;

		//If size of map reached, finishes
		if ((Addr) curC == ((Addr) elImage + sizeFile))
			break;
	}

	//Now that the information has been parsed, throws out
	//the image
	VG_(munmap) ( (void*)elImage, sizeFile );	

	//See what we got
//	printListRead();	
}

//GBS: Undoes masking at places that hold trampoline addresses
//This is if we have a supression list (passed in a file)
void RISE_(supress_instrumentation)(Addr start,Char *libname) {

	UChar *eip;
        libExcList *trav;
	oneAddress *adrs;

	eip = (UChar *)start;
	trav=headEL;
	while ((trav !=NULL) && (0 != VG_(strcmp)(libname,trav->libname))) {
		trav=trav->next;
	}
	if (trav != NULL) {
	       adrs=trav->aList;
	       while (adrs != NULL) {
		   //VG_(printf)("Suppressing : %x + %x (%x)\n",eip,adrs->eAdd,start+adrs->eAdd);
	           eip[adrs->eAdd] = eip[adrs->eAdd] ^ getMask(start+adrs->eAdd);
	           eip[adrs->eAdd+1] = eip[adrs->eAdd+1] ^ getMask(start+adrs->eAdd+1);
	           eip[adrs->eAdd+2] = eip[adrs->eAdd+2] ^ getMask(start+adrs->eAdd+2);
	           eip[adrs->eAdd+3] = eip[adrs->eAdd+3] ^ getMask(start+adrs->eAdd+3);
                   adrs = adrs->next;
                }
	}
}


//GBS: Output support for the in-section data reference patch
void RISE_(get_segment_name)(Addr refad, Addr content){
   encryptableSegment* trav;
   libExcList *travLibs, *newLib;
   oneAddress *adrs, *newAdd;
   Bool isNew;
   Int lenName;

   trav = encSegsHead;
   while ((trav!=NULL) && ((trav->start + trav->size -1) < refad))
                    trav = trav->next;

   travLibs=headEL;
   isNew = False;
   while ((travLibs !=NULL) && (0 != VG_(strcmp)(trav->fromFile,travLibs->libname))) {
		travLibs=travLibs->next;
   }
   if (travLibs != NULL) {
	       adrs=travLibs->aList;
	       while ((adrs != NULL) && (adrs->eAdd != refad)) {
                   adrs = adrs->next;
               }
	       if (adrs == NULL) {
			isNew = True;

			//Adds the number to the list for the library
			newAdd = (oneAddress *)VG_(arena_malloc)(VG_AR_CORE, sizeof(oneAddress));
			if (newAdd == NULL) {
                        	VG_(message)(Vg_UserMsg,"Error allocating memory for library info\n");
				VG_(exit)(1);
			}
			
			newAdd->eAdd = refad;
			newAdd->next = travLibs->aList;
			travLibs->aList = newAdd;
		}
   } else {
	isNew = True;
	//Create new library element
	newLib = (libExcList *)VG_(arena_malloc)(VG_AR_CORE, sizeof(libExcList));
	if (newLib ==NULL) {
                VG_(message)(Vg_UserMsg,"Error allocating memory for library info\n");
		VG_(exit)(1);
	}
	//Allocates enough memory for the name plus 0
	lenName = VG_(strlen)(trav->fromFile);
	newLib->libname = (Char *) VG_(arena_malloc)(VG_AR_CORE,lenName+1 );
	if (newLib->libname ==NULL) {
		VG_(message)(Vg_UserMsg,"Error allocating memory for library info\n");
		VG_(exit)(1);
	}

	//Copies the library name to the new location
	VG_(strncpy)(newLib->libname,trav->fromFile,lenName);
	newLib->libname[lenName]=0;
	newLib->next = NULL;
	newLib->aList = NULL;

	//Connects the new library info to the list
	if (headEL != NULL) 
		newLib->next = headEL;
	headEL = newLib;


	//Adds the number to the list for the library
	newAdd = (oneAddress *)VG_(arena_malloc)(VG_AR_CORE, sizeof(oneAddress));
	if (newAdd == NULL) {
                     	VG_(message)(Vg_UserMsg,"Error allocating memory for library info\n");
			VG_(exit)(1);
	}
			
	newAdd->eAdd = refad;
	newAdd->next = NULL;

	newLib->aList = newAdd;


   }

   if (isNew) {

   	VG_(printf)("inref,%s,",trav->fromFile);
   	//VG_(printf)("%x:%x\n",
   	//   refad - trav->sections[dotTEXT].start,content);
    	VG_(printf)("%x\n",
                   refad - trav->sections[dotTEXT].start);
                 return;
    }

}
