
/*--------------------------------------------------------------------*/
/*--- Memory-related stuff: segment initialisation and tracking,   ---*/
/*--- stack operations                                             ---*/
/*--- RISE: Management of auxiliary data structures with           ---*/
/*--- information about currently executable pieces                ---*/
/*---                                                  vg_memory.c ---*/
/*--------------------------------------------------------------------*/

/*
   This file is part of Valgrind, an extensible x86 protected-mode
   emulator for monitoring program execution on x86-Unixes.

   Copyright (C) 2000-2003 Julian Seward 
      jseward@acm.org

   Modified for RISE, an instruction set randomizer for security.

   Copyright (C) 2002-2003 Gabriela Barrantes
        gbarrant@ieee.org


   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.

   The GNU General Public License is contained in the file COPYING.
*/

#include "vg_include.h"
//#include "rise_scramble.c"
//#include "rise_magic_sequence_mgmnt.c"


//=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=
// - General utilities for memory-related operations
//=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=


//GBS: Return page aligned start address. This is to get
//the start of a memory block (page aligned)
//This function is duplicated here for conveniencd

Addr RISE_(getPageBasedStartOfBlock) ( Addr a )
{
   return a & ~(VKI_BYTES_PER_PAGE-1);
}


// GBS: Obtains Page aligned address given a non-page aligned address
// rounding up. This is to get the page boundary aligned end
// of a memory block
// Example: if page size = 4096, and block ends at address = 5079,
// then it will return 8192

Addr RISE_(getPageBasedEndOfBlock) ( Addr a )
{
        if ( (a & (VKI_BYTES_PER_PAGE-1)) == 0 )
        //Size is already page aligned
                return a;
        else
                return   (a & ~(VKI_BYTES_PER_PAGE-1)) + VKI_BYTES_PER_PAGE;
}

//=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=
//  -- Ordered Linked List (OLL) implementation of executable segment bookkeeping --
//  We are keeping a separate version even though Valgrind already has one listwith
//  some of this information
//=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=!=


//encryptableSegment and executableSection are
//defined on vg_include as it is used in vg_symtab
//main ponter to the list of executable segments
encryptableSegment* encSegsHead = NULL;

Bool RISE_(check_borders)(Addr eip) {
	int i;
       encryptableSegment *traverser;
       traverser = encSegsHead;
       while (traverser!=NULL) {
	     for (i=0;i<4;i++) {
	     	if ( (traverser->sections[i].valid) &&
		     (traverser->sections[i].start <= eip) &&
		     ( eip < (traverser->sections[i].size+traverser->sections[i].start) )) {
			return True;
		}
	     }
	     traverser = traverser->next;
       }
       return False;
		
}

//Adds a segment to the list (with empy information about sections)
encryptableSegment * RISE_(OLL_add_segment_to_list)(    Addr start, UInt size, Bool emulator,
                                                        Char *segName, Char rr,
Char ww, Char xx ) {

        encryptableSegment* newSegment, *traverser, *temp;
        Addr end;
        Int longName,i;

        //Trivial check
        if ((start == 0) || (size == 0))
          VG_(core_panic)("Attempted to register a segment of  size 0 or starting at address 0\n");

        //First check we are not overlapping -- I know it is
        //redundant with VG's check, but just in case...
        end = size+start-1;
        //VG_(printf)("Adding segment %x-%x\n",start,end);
        temp=NULL;
        traverser = encSegsHead;
        while (traverser != NULL) {
                //Checking for overlaps (ASSUMPTION: The list is ORDERED!)

                //VG_(printf)
                //("Comparing with segment %x-%x\n",traverser->start,traverser->start + traverser->size - 1);

                if ( ( end   < (traverser->start))
                                ||
                     ( start > ( (traverser->start) + (traverser->size) - 1))
                   ) {
                        //Non -overlapping

                        if ( traverser->start > start) break;
                        temp = traverser;
                        traverser = traverser->next;
                } else {
                        //Intersection!
                        VG_(core_panic)("Overlapping executable sections: cannot continue\n");
                }

        }

        //Create the new element...

        newSegment = (encryptableSegment *)VG_(arena_malloc)(VG_AR_CORE, sizeof(encryptableSegment));
        newSegment->start = start;
        newSegment->size = size;
        newSegment->emulator = emulator;

        longName = VG_(strlen)(segName)+1;
        newSegment->fromFile = (Char *) VG_(arena_malloc)(VG_AR_CORE, longName);        VG_(strncpy)(newSegment->fromFile,segName,longName-1);
        newSegment->fromFile[longName-1]=0;

        newSegment->rr = rr;
        newSegment->xx = xx;
        newSegment->ww = ww;

        for (i=0;i<numExeSec;i++) newSegment->sections[i].valid = False;

        newSegment->next = NULL;

        //And insert it (in order) in the list
        if (temp == NULL) { //inserting at the head
                if (encSegsHead != NULL)
                        newSegment->next = encSegsHead;
                encSegsHead =  newSegment;
        } else {
                newSegment->next = temp->next;
                temp->next = newSegment;
        }
        return (newSegment);

}

//Adds a section to a segment
void RISE_(OLL_add_section_to_segment)( encryptableSegment * theSegment, Addr start,
                                        UInt size, SectionType typeSec){
        Int i;
        Addr end;

        //Feeble attempt to ensure that the segment in question makes
        //some sort of sense
        if  (   (theSegment == NULL)
            || (start == 0)
            || (size == 0)
            ) VG_(core_panic)("ERROR: Trying to add section to null segment\n");
        //Do not allow overwrites over previous section data
        if (theSegment->sections[typeSec].valid)
                VG_(core_panic)("ERROR: Trying to rewrite an existing section\n");

        //Calculating the end of the section
        end = start + size -1;

        //Making sure that the section is inside the segment
        if (    (start < theSegment->start) ||
                ( end > (theSegment->start + theSegment->size - 1) )
           ) VG_(core_panic)("ERROR:Trying to add section that is outside the segment!\n");


        //Making sure that the section does not intersect with any of
        //the other sections
        for (i=0;i<numExeSec;i++) {
                if (theSegment->sections[i].valid)
                        if (    (end >= theSegment->sections[i].start)
                                                &&
                                (start  <= (theSegment->sections[i].start +
                                       theSegment->sections[i].size -1 ) )
                           ) //Overlapping sections!!!
                                VG_(core_panic)("ERROR:Trying to add overlapping section!\n");
        }

        //Actually adding the section
        theSegment->sections[typeSec].valid = True;
        theSegment->sections[typeSec].start = start;
        theSegment->sections[typeSec].size = size;
        theSegment->sections[typeSec].encrypted = False;
}


//GBS: If a section is eventually scrambled, this records the fact
void RISE_(noteEncryption)( encryptableSegment * theSegment, SectionType typeSec ) {
        theSegment->sections[typeSec].encrypted = True;
}
//GBS: A display helper
void print_section_info(encryptableSegment *theSeg,SectionType secTy) {
        VG_(printf)("Section ");
        switch (secTy) {
                case dotINIT: VG_(printf)(" INIT ");
                              break;
                case dotTEXT: VG_(printf)(" TEXT ");
                              break;
                case dotPLT : VG_(printf)(" PLT ");
                              break;
                case dotFINI: VG_(printf)(" FINI ");
                              break;
                default: VG_(printf)(" Undefined type of section ");
        }
        VG_(printf)("\n");
        if (theSeg->sections[secTy].valid) {
                VG_(printf)("VALID!\n");
                VG_(printf)("\t[%x-%x]\n",theSeg->sections[secTy].start,
                        theSeg->sections[secTy].start+
                        theSeg->sections[secTy].size-1);
                if (theSeg->sections[secTy].encrypted)
                        VG_(printf)("ALREADY ENCRYPTED\n");
                else
                        VG_(printf)("Non encrypted!\n");
        }
        else
                VG_(printf)("Non valid!\n");
}


//GBS: Another display function
void print_segment_info(encryptableSegment *theSeg) {
        VG_(printf)("Segment [%x-%x]\n",theSeg->start,theSeg->start+theSeg->size-1);
        if (theSeg->emulator) VG_(printf)("EMULATOR segment\n");
        else VG_(printf)("Normal segment\n");
        VG_(printf)("Name of library %s\n",theSeg->fromFile);
        VG_(printf)("Sections:\n");
        print_section_info(theSeg,dotINIT);
        print_section_info(theSeg,dotPLT);
        print_section_info(theSeg,dotTEXT);
        print_section_info(theSeg,dotFINI);
}

//GBS: RISE specifics of segment removal while the process
//is still executing. It does happen ==> libraries get unmapped
//while a process is alive.

void removeAllPieces(Addr start) {
/* GBS: Be very careful when modifying this because
 * we don't store all information anymore, only some
 * pieces */
	

   encryptableSegment* trav, *temp;

  // VG_(printf)("REMV: Removing segment starting at %x\n",start);

   temp = NULL;
   trav = encSegsHead;
   while ( (trav != NULL) && (trav->start < start) ) {
           temp = trav;
           trav = trav->next;
   }

   //Sanity check
   if ((trav == NULL) || (trav->emulator) || (trav->start != start))
           VG_(core_panic)("Error in segment list\n");

   //Release the space taken by the name of the file
   VG_(arena_free)(VG_AR_CORE,trav->fromFile);

  // VG_(printf)("REMV: Removing piece %x-%x\n",trav->start,trav->start + trav->size -1);

   //Now, regenerate the key space (just in case)
   //regenerateKey(trav->start,trav->size);

   //And now, release the space taken by the record
   if (trav != encSegsHead)
        temp->next = trav->next;
   else
        encSegsHead=trav->next;
   VG_(arena_free)(VG_AR_CORE,trav);


}

//----------------------------------------******-----------------------------------
//GBS: For the patch around libraries using data inside what should
//be strictly code: This is instrumentation that at run time
//initializes the sections limits so they could be checked when
//a dangerous instruction is executed
//It gets called only if a potentially dangerous addressing form
//is found in a basic block
//If the segment to which the address belongs is not found, it should
//crash horribly because
//in legal circumstances it is invoked from real code

Addr llStart = 0;
Addr llEnd = 0;

void loadLimits(Addr refad, UCodeBlock * cb, Addr *start, Addr *end) {
//This is way heavy handed, but incredibly there are applications
//that use trampolines in init and even in fini

   encryptableSegment* trav;
   Addr endsec;

   //Just to avoid checking the list every time
   if ((refad >= llStart) && (refad <= llEnd)) {
           // LIMIT ERRROR!!
           //VG_(printf)("LIMITS (%x) in %x-%x\n",
           //              refad, llStart, llEnd);
           VG_(set_global_var)(cb, (Addr) start, llStart);
           VG_(set_global_var)(cb, (Addr) end, llEnd);
           return;
   }
   trav = encSegsHead;
   //First find the segment
   while ( (trav !=NULL) && ( (trav->start + trav->size -1 ) < refad) ){
           trav=trav->next;
   }

   //The segment must ALWAYS be found because the address is plucked from
   //currently executing code
   if ( (trav==NULL) || (trav->start > refad))
     VG_(core_panic)("(Inreference tracking) A supposedly executable address in non-exec space (1)\n");

   //Internal references can happen in init/text/fini sections but not
   //plt because it has a standardized form that doesn't allow for inreferences
   //The following is just a sanity check
    if  ( !(trav->sections[dotTEXT].valid)) {
           VG_(message)(Vg_UserMsg,"Segment %s without text section\n",trav->fromFile);
           exit(1);
    }

    //is this address in .text?
    endsec = trav->sections[dotTEXT].start + trav->sections[dotTEXT].size - 1;
    if    ((refad >= trav->sections[dotTEXT].start) && (refad <= endsec) )  {
                //LIMIT ERROR TRACING!
                //VG_(printf)("LIMITS (%x) in (%x-%x) (.text) of %s\n",
                //      refad,trav->sections[dotTEXT].start,endsec,trav->fromFile);
                VG_(set_global_var)(cb, (Addr) start, trav->sections[dotTEXT].start);
                VG_(set_global_var)(cb, (Addr) end, endsec);
                llStart = trav->sections[dotTEXT].start;
                llEnd = endsec;
                return;
    }

    //or maybe in .init?
    if  ( trav->sections[dotINIT].valid) {
        endsec = trav->sections[dotINIT].start + trav->sections[dotINIT].size -
1;
        if ( (refad >= trav->sections[dotINIT].start) &&  (refad <= endsec) )  {                //LIMIT ERROR TRACING!
                //VG_(printf)("LIMITS (%x) in %x-%x (.init) of %s\n",
                //      refad,trav->sections[dotINIT].start,endsec,trav->fromFile);
                VG_(set_global_var)(cb, (Addr) start, trav->sections[dotINIT].start);
                VG_(set_global_var)(cb, (Addr) end, endsec);
                llStart = trav->sections[dotINIT].start;
                llEnd = endsec;
                return;
        }
    }

    //.fini?
    if  ( trav->sections[dotFINI].valid) {
        endsec = trav->sections[dotFINI].start + trav->sections[dotFINI].size -
1;
        if ((refad >= trav->sections[dotFINI].start) &&  (refad <= endsec) )  {
                //LIMIT ERROR TRACING!
                //VG_(printf)("LIMITS (%x) in %x-%x (.fini) of %s\n",
                //      refad,trav->sections[dotFINI].start,endsec,trav->fromFile);
                VG_(set_global_var)(cb, (Addr) start, trav->sections[dotFINI].start);
                VG_(set_global_var)(cb, (Addr) end, endsec);
                llStart = trav->sections[dotFINI].start;
                llEnd = endsec;
                return;
        }
    }

     VG_(core_panic)("(Inreference Tracking) A supposedly executable address in non-exec space (2)\n");
}

// GBS??
//This is just a display flag for the family of debug messages
//relative to the elf analysis (in theory). In practice it is
//being used for most error messages
Bool mapinfo = False;


/*--------------------------------------------------------------*/
/*--- Original Valgrind stuff                                ---*/
/*--------------------------------------------------------------*/


/* Define to debug the memory-leak-detector. */
/* #define VG_DEBUG_LEAKCHECK */


/*--------------------------------------------------------------*/
/*--- Initialise program data/text etc on program startup.   ---*/
/*--------------------------------------------------------------*/

typedef
   struct _ExeSeg {
      Addr start;
      UInt size;
      struct _ExeSeg* next;
   }
   ExeSeg;

/* The list of current executable segments loaded.  Required so that when a
   segment is munmap'd, if it's executable we can recognise it as such and
   invalidate translations for it, and drop any basic-block specific
   information being stored.  If symbols are being used, this list will have
   the same segments recorded in it as the SegInfo symbols list (but much
   less information about each segment).
*/
static ExeSeg* exeSegsHead = NULL;


//=============================================================================
//GBS: Using the info stored in currentSeg, scrambles the relevant areas

void RISE_(scrambleLibrary)(encryptableSegment * currentSeg) {

UInt i,j,baseToMprotect,offsetToMprotect;
Bool rightsChange;
Addr start;
UInt size;
Char xx,rr,ww;
Int prot;
UChar * eip;

        //If this function is invoked from an emulator area
	//just ignore it

        if (currentSeg->emulator) return;

        //library rights
        xx = currentSeg->xx;
        rr = currentSeg->rr;
        ww = currentSeg->ww;

        for (i=0;i<numExeSec;i++) {

                if (currentSeg->sections[i].valid &&
                    !(currentSeg->sections[i].encrypted)) {

                        rightsChange = False;

                        //Section limits
                        start = currentSeg->sections[i].start;
                        size  = currentSeg->sections[i].size;

                        //Start of block (page aligned)
                        baseToMprotect = RISE_(getPageBasedStartOfBlock)(start);                        //End of block (page aligned)
                        offsetToMprotect = RISE_(getPageBasedEndOfBlock)(start+size-1) - baseToMprotect;

                        if ((rr != 'r') || (ww != 'w')) {
                                prot = VKI_PROT_READ|VKI_PROT_WRITE;
                                VG_(mprotect)((void *)baseToMprotect,offsetToMprotect,prot);
                                rightsChange = True;
                        }

                        //VG_(printf)("Encoding section %x-%x\n",start,start+size-1);

                        eip = (UChar *) start;

                        //Here's the actual "encryption" process
                        for (j=0;j<size;j++) {
                                eip[j] = eip[j] ^ getMask(start + j);
                        }


                        //Returns segment to its original permissions
                        if ( rightsChange ) {
                                prot = VKI_PROT_EXEC;
                                if (rr == 'r') prot = prot|VKI_PROT_READ;
                                if (ww == 'w') prot = prot|VKI_PROT_WRITE;
                                VG_(mprotect)((void *)baseToMprotect,offsetToMprotect,prot);
                        }

                        //Marks the segment as already (physically) masked
                        RISE_(noteEncryption)( currentSeg, i );

                }
        }

        //If a suppression list is given, decrypts the requested bytes
        //As of now, it assumes absolute displacement from start of library
        if ( RISE_(noInstrumentation) ) {
                RISE_(supress_instrumentation)(currentSeg->start,currentSeg->fromFile);
        }
}

/* DEFINED NOW IN rise_magic_sequence_mgmnt.c
 * void RISE_(register_magic_area)(Addr start, UInt size)
 */

/* Prepend it -- mmaps/munmaps likely to follow a stack pattern(?) so this
   is good.
   Also check no segments overlap, which would be very bad.  Check is linear
   for each seg added (quadratic overall) but the total number should be
   small (konqueror has around 50 --njn). */
static void add_exe_segment_to_list( Addr a, UInt len )
{
   Addr lo = a;
   Addr hi = a + len - 1;
   ExeSeg* es;
   ExeSeg* es2;

   /* Prepend it */
   es        = (ExeSeg*)VG_(arena_malloc)(VG_AR_CORE, sizeof(ExeSeg));
   es->start = a;
   es->size  = len;
   es->next  = exeSegsHead;
   exeSegsHead = es;

   /* Check there's no overlap with the rest of the list */
   for (es2 = es->next; es2 != NULL; es2 = es2->next) {
      Addr lo2 = es2->start;
      Addr hi2 = es2->start + es2->size - 1;
      Bool overlap;
      vg_assert(lo < hi);
      vg_assert(lo2 < hi2);
      /* the main assertion */
      overlap = (lo <= lo2 && lo2 <= hi)
                 || (lo <= hi2 && hi2 <= hi);
      if (overlap) {
         VG_(printf)("\n\nOVERLAPPING EXE SEGMENTS\n"
                     "  new: start %p, size %d\n"
                     "  old: start %p, size %d\n\n",
                     es->start, es->size, es2->start, es2->size );
         vg_assert(! overlap);
      }
   }
}

static Bool remove_if_exe_segment_from_list( Addr a )
{
   ExeSeg **prev_next_ptr = & exeSegsHead,
          *curr = exeSegsHead;

   while (True) {
      if (curr == NULL) break;
      if (a == curr->start) break;
      prev_next_ptr = &curr->next;
      curr = curr->next;
   }
   if (curr == NULL)
      return False;

   // GBS: RISE stuff has to be removed as well
   removeAllPieces(a);

   vg_assert(*prev_next_ptr == curr);

   *prev_next_ptr = curr->next;

   VG_(arena_free)(VG_AR_CORE, curr);
   return True;
}

/* Records the exe segment in the ExeSeg list (checking for overlaps), and
   reads debug info if required.  Note the entire /proc/pid/maps file is 
   read for the debug info, but it just reads symbols for newly added exe
   segments.  This is required to find out their names if they have one.  So
   we don't use this at startup because it's overkill and can screw reading
   of /proc/pid/maps.
 */
void VG_(new_exe_segment) ( Addr a, UInt len )
{
   add_exe_segment_to_list( a, len );
   //GBS: This is duplicate work, but seems easy to
   //mantain this way
   VG_(read_symbols)();
}

/* Invalidate translations as necessary (also discarding any basic
   block-specific info retained by the skin) and unload any debug
   symbols. */
// Nb: remove_if_exe_segment_from_list() and VG_(maybe_unload_symbols)()
// both ignore 'len', but that seems that's ok for most programs...  see
// comment above vg_syscalls.c:mmap_segment() et al for more details.
void VG_(remove_if_exe_segment) ( Addr a, UInt len )
{
   if (remove_if_exe_segment_from_list( a )) {
      VG_(invalidate_translations) ( a, len, False );
      VG_(unload_symbols)          ( a, len );
   }
}

static
void startup_segment_callback ( Addr start, UInt size,
                                Char rr, Char ww, Char xx,
                                UInt foffset, UChar* filename )
{
   UInt r_esp;
   Bool is_stack_segment;
   Bool verbose = False; /* set to True for debugging */

   if (verbose)
      VG_(message)(Vg_DebugMsg,
                   "initial map %8x-%8x %c%c%c? %8x (%d) (%s)",
                   start,start+size,rr,ww,xx,foffset,
                   size, filename?filename:(UChar*)"NULL");

   if (rr != 'r' && xx != 'x' && ww != 'w') {
      /* Implausible as it seems, R H 6.2 generates such segments:
      40067000-400ac000 r-xp 00000000 08:05 320686 /usr/X11R6/lib/libXt.so.6.0
      400ac000-400ad000 ---p 00045000 08:05 320686 /usr/X11R6/lib/libXt.so.6.0
      400ad000-400b0000 rw-p 00045000 08:05 320686 /usr/X11R6/lib/libXt.so.6.0
      when running xedit. So just ignore them. */
      if (0)
         VG_(printf)("No permissions on a segment mapped from %s\n",
                     filename?filename:(UChar*)"NULL");
      return;
   }

   /* If this segment corresponds to something mmap'd /dev/zero by the
      low-level memory manager (vg_malloc2.c), skip it.  Clients
      should never have access to the segments which hold valgrind
      internal data.  And access to client data in the VG_AR_CLIENT
      arena is mediated by the skin, so we don't want make it
      accessible at this stage. */
   if (VG_(is_inside_segment_mmapd_by_low_level_MM)( start )) {
      if (verbose)
         VG_(message)(Vg_DebugMsg,
                      "   skipping %8x-%8x (owned by our MM)",
                      start, start+size );
      /* Don't announce it to the skin. */
      return;
   }

   /* This parallels what happens when we mmap some new memory */
   if (filename != NULL && xx == 'x') {
      VG_(new_exe_segment)( start, size );
      //GBS: additionally, we go and encrypt the segment
      //(NONEED?)RISE_(read_symtab_callback)(start,size,rr,ww,xx,foffset,filename);
   }

   VG_TRACK( new_mem_startup, start, size, rr=='r', ww=='w', xx=='x' );

   /* If this is the stack segment mark all below %esp as noaccess. */
   r_esp = VG_(baseBlock)[VGOFF_(m_esp)];
   is_stack_segment = start <= r_esp && r_esp < start+size;
   if (is_stack_segment) {
      if (0)
         VG_(message)(Vg_DebugMsg, "invalidating stack area: %x .. %x",
                      start,r_esp);
      VG_TRACK( die_mem_stack, start, r_esp-start );
   }
}


/* 1. Records startup segments from /proc/pid/maps.  Takes special note
      of the executable ones, because if they're munmap()ed we need to
      discard translations.  Also checks there's no exe segment overlaps.

      Note that `read_from_file' is false;  we read /proc/self/maps into a
      buffer at the start of VG_(main) so that any superblocks mmap'd by
      calls to VG_(malloc)() by SK_({pre,post}_clo_init) aren't erroneously
      thought of as being owned by the client.

   2. Sets up the end of the data segment so that vg_syscalls.c can make
      sense of calls to brk().
 */
void VG_(init_memory) ( void )
{

  //GBS: inits mask cache
  RISE_(init_level0)();

  //GBS: inits the structure of authorized Magic Sequence users
  RISE_(init_magic_users)();

   /* 1 */
   VG_(read_procselfmaps) ( startup_segment_callback, /*read_from_file*/False );

   /* 2 */
   VG_(init_dataseg_end_for_brk)();

   /* kludge: some newer kernels place a "sysinfo" page up high, with
      vsyscalls in it, and possibly some other stuff in the future. */
   if (VG_(sysinfo_page_exists)) {
      VG_(new_exe_segment)( VG_(sysinfo_page_addr), 4096 );
      VG_TRACK( new_mem_startup, VG_(sysinfo_page_addr), 4096, 
                True, True, True );
     }
}

/*------------------------------------------------------------*/
/*--- Tracking permissions around %esp changes.            ---*/
/*------------------------------------------------------------*/

/*
   The stack
   ~~~~~~~~~
   The stack's segment seems to be dynamically extended downwards
   by the kernel as the stack pointer moves down.  Initially, a
   1-page (4k) stack is allocated.  When %esp moves below that for
   the first time, presumably a page fault occurs.  The kernel
   detects that the faulting address is in the range from %esp upwards
   to the current valid stack.  It then extends the stack segment
   downwards for enough to cover the faulting address, and resumes
   the process (invisibly).  The process is unaware of any of this.

   That means that Valgrind can't spot when the stack segment is
   being extended.  Fortunately, we want to precisely and continuously
   update stack permissions around %esp, so we need to spot all
   writes to %esp anyway.

   The deal is: when %esp is assigned a lower value, the stack is
   being extended.  Create a secondary maps to fill in any holes
   between the old stack ptr and this one, if necessary.  Then 
   mark all bytes in the area just "uncovered" by this %esp change
   as write-only.

   When %esp goes back up, mark the area receded over as unreadable
   and unwritable.

   Just to record the %esp boundary conditions somewhere convenient:
   %esp always points to the lowest live byte in the stack.  All
   addresses below %esp are not live; those at and above it are.  
*/

/* Kludgey ... how much does %esp have to change before we reckon that
   the application is switching stacks ? */
#define VG_PLAUSIBLE_STACK_SIZE  8000000
//GBS: umoddi/udivdi adjustment (no change!)
//#define VG_HUGE_DELTA            (VG_PLAUSIBLE_STACK_SIZE / 4)
#define VG_HUGE_DELTA            (VG_PLAUSIBLE_STACK_SIZE / 4)

/* This function gets called if new_mem_stack and/or die_mem_stack are
   tracked by the skin, and one of the specialised cases (eg. new_mem_stack_4)
   isn't used in preference */
__attribute__((regparm(1)))
void VG_(unknown_esp_update)(Addr new_ESP)
{
   Addr old_ESP = VG_(get_archreg)(R_ESP);
   Int  delta   = (Int)new_ESP - (Int)old_ESP;

   if (delta < -(VG_HUGE_DELTA) || VG_HUGE_DELTA < delta) {
      /* %esp has changed by more than HUGE_DELTA.  We take this to mean
         that the application is switching to a new stack, for whatever
         reason. 
       
         JRS 20021001: following discussions with John Regehr, if a stack
         switch happens, it seems best not to mess at all with memory
         permissions.  Seems to work well with Netscape 4.X.  Really the
         only remaining difficulty is knowing exactly when a stack switch is
         happening. */
      if (VG_(clo_verbosity) > 1)
           VG_(message)(Vg_UserMsg, "Warning: client switching stacks?  "
                                    "%%esp: %p --> %p", old_ESP, new_ESP);
   } else if (delta < 0) {
      VG_TRACK( new_mem_stack, new_ESP, -delta );

   } else if (delta > 0) {
      VG_TRACK( die_mem_stack, old_ESP,  delta );
   }
}

/*--------------------------------------------------------------------*/
/*--- end                                              vg_memory.c ---*/
/*--------------------------------------------------------------------*/

